/***
*
* Copyright (C) 2006,2007,2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>

#include "mex.h"
#include "dream_error.h"

#define TRUE 1
#define FALSE 0

#include "dream.h"

/***
 *
 *  Parallel (threaded) matrix and sub-matrix copy.
 *
 ***/

//
// Globals
//

volatile int running;

//
// typedef:s
//

typedef struct
{
  dream_idx_type line_start;
  dream_idx_type line_stop;
  double *A;
  dream_idx_type A_M;
  dream_idx_type A_N;
  double *B;
  dream_idx_type B_M;
  dream_idx_type B_N;
  dream_idx_type r;
  dream_idx_type k;
  dream_idx_type len;
} DATA;

typedef void (*sighandler_t)(int);

//
// Function prototypes.
//
void* smp_process(void *arg);
void sighandler(int signum);
void sig_abrt_handler(int signum);
void sig_keyint_handler(int signum);

/***
 *
 * Thread function. 
 *
 ***/

void* smp_process(void *arg)
{
  DATA D = *(DATA *)arg;
  dream_idx_type    line_start=D.line_start, line_stop=D.line_stop, n;
  double *A = (double*) D.A, *B = D.B;
  dream_idx_type A_M = D.A_M, B_M = D.B_M, r = D.r, k = D.k, len = D.len;

  for (n=line_start; n<line_stop; n++) {
    
    memcpy( &A[r+(k+n)*A_M], &B[0+n*B_M], len*sizeof(double));
    
    if (running==FALSE) {
      printf("copy_p: thread for column %d -> %d bailing out!\n",line_start+1,line_stop);
      break;
    }
 
  }
  return(NULL);
}



/***
 *
 * Signal handlers.
 *
 ***/

void sighandler(int signum) {
  //printf("Caught signal SIGTERM.\n");
  running = FALSE;
}

void sig_abrt_handler(int signum) {
  //printf("Caught signal SIGABRT.\n");
}

void sig_keyint_handler(int signum) {
  //printf("Caught signal SIGINT.\n");
}

/***
 * 
 * Matlab (MEX) gateway function for copy_p.
 *
 ***/

void  mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  double         *A,*B; 
  sighandler_t   old_handler, old_handler_abrt, old_handler_keyint;
  pthread_t      *threads;
  dream_idx_type line_start, line_stop, A_M, A_N, B_M, B_N;
  int            thread_n, N, err;
  dream_idx_type r_M, r_N, k_M, k_N;
  double         *r, *k;
  void           *retval;
  DATA           *D;

  // Check for proper number of arguments


  if (nrhs != 5)
    mexErrMsgTxt("copy_p requires 5 input arguments!");

  if (nlhs > 0)
    mexErrMsgTxt("Too many output arguments for copy_p, none required!");
  
  A_M = mxGetM(prhs[0]);
  A_N = mxGetN(prhs[0]);
  A   = mxGetPr(prhs[0]);

  r_M = mxGetM(prhs[1]);
  r_N = mxGetN(prhs[1]);
  r   = mxGetPr(prhs[1]);

  k_M = mxGetM(prhs[2]);
  k_N = mxGetN(prhs[2]);
  k   = mxGetPr(prhs[2]);

  B_M = mxGetM(prhs[3]);
  B_N = mxGetN(prhs[3]);
  B   = mxGetPr(prhs[3]);

  // Check that arg 2. 
  if ( r_M * r_N !=2 )
    mexErrMsgTxt("Argument 2 must be a 2 element vector!");

  if ( (dream_idx_type) r[1] < 1)
    mexErrMsgTxt("1st element of argument 2 must be > 1!");

  if (( (dream_idx_type) r[0] > A_M) || ( (dream_idx_type) r[1] > A_M))
    mexErrMsgTxt("One element of argument 2 exceeds row dimension of arg 1!");

  // Check that arg 3. 
  if ( k_M * k_N !=2 )
    mexErrMsgTxt("Argument 3 must be a 2 element vector!");

  if ( (dream_idx_type) k[1] < 1)
    mexErrMsgTxt("1st element of argument 3 must be > 1!");
  
  if (( (dream_idx_type) k[0] > A_N) || ( (dream_idx_type) k[1] > A_N))
    mexErrMsgTxt("One element of argument 3 exceeds colomn dimension of arg 1!");

  // Check that arg 4. 
  if (B_M < (dream_idx_type) (r[1]-r[0])+1 )
    mexErrMsgTxt("Argument 4 has not the number of rows as given by arg 2!");

  if (B_N != (dream_idx_type) (k[1]-k[0])+1 )
    mexErrMsgTxt("Argument 4 has not the number of columns as given by arg 3!");

    
  //
  // Number of threads.
  //

  // Check that arg 5 is a scalar.
  if ( mxGetM(prhs[4]) * mxGetN(prhs[4]) !=1)
    mexErrMsgTxt("Argument 5 must be a scalar!");
  
  N = (int) mxGetScalar(prhs[2]);
  if (N < 1)
    mexErrMsgTxt("Argument 5 must be larger or equal to 1 (min number of CPUs)!");

  // Check n_cpus argument.
  if (N > A_N) {
    // Add a -v verbose flag to display stuff like this!
    //mexPrintf("Warning: n_cpus is larger then number of columns in first arg.\n");
    //mexPrintf("         Setting n_cpus = # cols in 1st arg!\n");
    N = A_N; //
  }

  
  
  //
  // Register signal handlers.
  //

  if ((old_handler = signal(SIGTERM, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }

  if ((old_handler_abrt=signal(SIGABRT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  if ((old_handler_keyint=signal(SIGINT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  //
  // Call the mem_copy subroutine.
  //

  running = TRUE;
  
  // Allocate local data.
  D = (DATA*) malloc(N*sizeof(DATA));
  if (!D)
    mexErrMsgTxt("Failed to allocate memory for thread data!");

  // Allocate mem for the threads.
  threads = (pthread_t*) malloc(N*sizeof(pthread_t));
  
  if (!threads)
    mexErrMsgTxt("Failed to allocate memory for threads!");

  for (thread_n = 0; thread_n < N; thread_n++) {

    line_start = thread_n * A_N/N;
    line_stop =  (thread_n+1) * A_N/N;

    // Init local data.
    D[thread_n].line_start = line_start; // Local start index;
    D[thread_n].line_stop = line_stop; // Local stop index;
    D[thread_n].A = A;
    D[thread_n].A_M = A_M;
    D[thread_n].A_N = A_N;
    D[thread_n].B = B;
    D[thread_n].B_M = B_M;
    D[thread_n].B_N = B_N;
    D[thread_n].r = (dream_idx_type) (r[0]-1);
    D[thread_n].k = (dream_idx_type) (k[0]-1);
    D[thread_n].len = (dream_idx_type) (r[1]-r[0]+1);

    // Starts the threads.
    err = pthread_create(&threads[thread_n], NULL, smp_process, &D[thread_n]);
    if (err != 0)
      mexErrMsgTxt("Error when creating a new thread!\n");
  }
  
  // Wait for all threads to finish.
  for (thread_n = 0; thread_n < N; thread_n++) {
    err = pthread_join(threads[thread_n], &retval);
    if (err != 0)
      mexErrMsgTxt("Error when joining a thread!\n");
  }


  // Free memory.
  if (D) {
    free((void*) D);
  }

  //
  // Restore old signal handlers.
  //

  if (signal(SIGTERM, old_handler) == SIG_ERR) {
    printf("Couldn't register old signal handler.\n");
  }
   
  if (signal(SIGABRT,  old_handler_abrt) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  if (signal(SIGINT, old_handler_keyint) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }

  if (!running) {
    // This may kill Matlab 7 release 14 (bug in Matlab).
    mexErrMsgTxt("CTRL-C pressed!\n"); // Bail out.
  }
  
  if (threads) {
    free((void*) threads);
  }

  return;
}
