/***
*
* Copyright (C) 2003,2004,2005,2006,2007,2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "mex.h"
#include "dream_error.h"

/***
 *
 * Time-domain (monostatic) Synthetic Aperture Focusing Technique (SAFT)
 *
 ***/

#if defined(_MSC_VER) || defined(__LCC__)
#include "msvc_rint.h"
#endif

/***
 * 
 * Matlab (MEX) gateway function for SAFT.
 *
 ***/

void  mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  int n, k_shift, l;
  int K, L, no, n_l;
  double *s_par, *m_par;
  double *B, *Bsaft, z_prim, cp , a, dt, d, *ro, xo, yo, zo, r_xy;
  double *r_trans, x_trans, y_trans, z_trans, *delay;
  double z, tmp;
  
  // Check for proper number of arguments
  
  if (nrhs != 7) {
    mexErrMsgTxt("saft requires 7 input arguments!");
  }
  else
    if (nlhs > 1) {
      mexErrMsgTxt("Too many output arguments for saft!");
    }
  
  //
  // B-scan matrix.
  //

  K = mxGetM(prhs[0]); // Number of temporal samples.
  L = mxGetN(prhs[0]); // Number of transducer positions (A-scans).
  B = mxGetPr(prhs[0]);

  //
  // Transducer position matrix.
  //

  // Check that arg 2 is a L x 3 is a matrix.
  if (!(mxGetM(prhs[1])==L && mxGetN(prhs[1])==3))
    mexErrMsgTxt("Arg 2 must be a (number of transducer positions) x 3 matrix!");

  r_trans = mxGetPr(prhs[1]);

  //
  // Start point of impulse response vector ([us]).
  //

  if ( (mxGetM(prhs[2]) * mxGetN(prhs[2]) !=1) && ((mxGetM(prhs[2]) * mxGetN(prhs[2])) != L))
    dream_err_msg("Argument 3 (delay(s)) must be a scalar or a vector with a length equal to the number of A-scans!");
  
  delay = mxGetPr(prhs[2]);

  //
  // Sampling parameter.
  //

  // Check that arg 4 is a scalar.
  if (!((mxGetM(prhs[3])==1 && mxGetN(prhs[3])==1)))
    dream_err_msg("Argument 4 (temporal sampling period) must be a scalar");

  s_par = mxGetPr(prhs[3]);
  dt    = s_par[0]; // Temporal discretization size (= 1/sampling freq).

  //
  // Material parameter.
  //

  // Check that arg 5 is a scalar.
  if (!((mxGetM(prhs[4])==1 && mxGetN(prhs[4])==1)))
    dream_err_msg("Argument 5 (sound speed) must be a scalar!");
  
  m_par = mxGetPr(prhs[4]);
  cp    = m_par[0]; // Sound speed.

  //
  // Observation point matrix.
  //

  // Check that arg 6 is a (number of observation points) x 3 matrix.
  if (!mxGetN(prhs[5])==3)
    dream_err_msg("Argument 6 must be a (number of observation points) x 3 matrix!");

  no = mxGetM(prhs[5]); // Number of observation points.
  ro = mxGetPr(prhs[5]);

  //
  // Synthetic Aperture
  //
 
  // Check that arg 7 is scalar.
  if ((mxGetM(prhs[6])!=1) && (mxGetN(prhs[6])!=1))
    mexErrMsgTxt("Argument 7 must be a scalar!");

  a = mxGetScalar(prhs[6]);

  //
  // Create an output matrix for the procssed image.
  //

  plhs[0] = mxCreateDoubleMatrix(no,1,mxREAL);
  Bsaft = mxGetPr(plhs[0]);

  //
  // SAFT processing with linear interpolation.
  //

  // Loop over all observation points.
  for (n=0; n<no; n++) {
    
    // Current observation point.
    xo = ro[n]; 
    yo = ro[n+1*no]; 
    zo = ro[n+2*no]; 
    
    Bsaft[n] = 0.0;
    
    // Loop over elements in the aperture.
    n_l = 0;
    for (l=0; l<L; l++) {
      
      x_trans = r_trans[l];
      y_trans = r_trans[l+1*L];
      z_trans = r_trans[l+2*L];
	
      // Horizontal distance between transducer and observation point.
      r_xy = sqrt( (x_trans - xo)*(x_trans - xo) + (y_trans - yo)*(y_trans - yo) );

      // if r_xy is inside the synthetic aperture.
      if (r_xy <= a/2) {
	n_l ++;

	// Vertical distance between transducer and observation point.
	z = zo - z_trans;
	  
	z_prim = sqrt(z*z + r_xy*r_xy);
	  
	// Sampling index at depth z (double path).
	//k = rint(  z/cp/dt*2*1e3 - us[0]/dt );
	  
	// Focusing shift.
	//shift = rint( 2*(z_prim - z)/cp/dt * 1e3 );
	  
	// Rounding err.
	//d = 2*(z_prim - z)/cp/dt * 1e3 - ((double) shift);
	  
	// Shifted sample index.
	//k_shift = k + shift;
	  
	// Temp. storage to avoid recomputing.
	//tmp = z/cp/dt*2*1e3 - us[0]/dt + 2*(z_prim - z)/cp/dt * 1e3;

	if (mxGetM(prhs[2]) * mxGetN(prhs[2]) == 1) 
	  tmp = (2*z_prim/cp * 1e3 - 2*delay[0])/dt; 
	else
	  tmp = (2*z_prim/cp * 1e3 - 2*delay[n])/dt; 
	  
	// Better to round just one time!
	k_shift = (int) rint(tmp);

	// Rounding err.
	d =  tmp - ((double) k_shift);
	  
	// Linear interpolation.
	if ((k_shift+1 < K) && (k_shift-1 > 0) ) {
	  if (d >=0) {
	    Bsaft[n] += (1.0 - d) * B[k_shift     + l*K];
	    Bsaft[n] += d * B[(k_shift+1) + l*K];
	  }
	  else {
	    Bsaft[n] -= d * B[k_shift     + l*K];
	    Bsaft[n] += (1.0 + d) * B[(k_shift-1) + l*K];
	  }
	} // if
      } // SAFT loop. 
	
      //Bsaft[n] /= (double) n_l; // Normalize with # elements in aperture.
    }
  }
  
  return;
}

  
