/***
*
* Copyright (C) 2006,2007,2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>

//#include "dream_error.h"

//
// Octave headers.
//

#include <octave/oct.h>

#include <octave/config.h>

#include <iostream>
using namespace std;

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

#define TRUE 1
#define FALSE 0

#include "dream.h"

/***
 *
 *  Parallel (threaded) matrix and sub-matrix copy.
 *
 ***/

//
// Globals
//

volatile int running;

//
// typedef:s
//

typedef struct
{
  dream_idx_type line_start;
  dream_idx_type line_stop;
  double *A;
  dream_idx_type A_M;
  dream_idx_type A_N;
  double *B;
  dream_idx_type B_M;
  dream_idx_type B_N;
  dream_idx_type r;
  dream_idx_type k;
  dream_idx_type len;
} DATA;

typedef void (*sighandler_t)(int);

//
// Function prototypes.
//
void* smp_process(void *arg);
void sighandler(int signum);
void sig_abrt_handler(int signum);
void sig_keyint_handler(int signum);

/***
 *
 * Thread function. 
 *
 ***/

void* smp_process(void *arg)
{
  DATA D = *(DATA *)arg;
  dream_idx_type    line_start=D.line_start, line_stop=D.line_stop, n;
  double *A = (double*) D.A, *B = D.B;
  dream_idx_type A_M = D.A_M, B_M = D.B_M, r = D.r, k = D.k, len = D.len;

  for (n=line_start; n<line_stop; n++) {
    
    memcpy( &A[r+(k+n)*A_M], &B[0+n*B_M], len*sizeof(double));
    
    if (running==FALSE) {
      printf("copy_p: thread for column %d -> %d bailing out!\n",line_start+1,line_stop);
      break;
    }
 
  }
  return(NULL);
}



/***
 *
 * Signal handlers.
 *
 ***/

void sighandler(int signum) {
  //printf("Caught signal SIGTERM.\n");
  running = FALSE;
}

void sig_abrt_handler(int signum) {
  //printf("Caught signal SIGABRT.\n");
}

void sig_keyint_handler(int signum) {
  //printf("Caught signal SIGINT.\n");
}

/***
 * 
 * Octave (OCT) gateway function for copy_p.
 *
 ***/

DEFUN_DLD (copy_p, args, nlhs,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {} copy_p(A,col_idx,row_idx,B,n_cpus);\n\
\n\
COPY_P Performs parallel (inplace) copy of the matrix B into the matrix A using threads .\n\
\n\
Input parameters:\n\
\n\
@table @samp\n\
@item A\n\
An MxN matrix.\n\
@item row_idx\n\
row_idx is a two element vector where row_idx(1) is the start index \n\
@item col_idx\n\
col_idx is a two element vector where col_idx(1) is the start index \n\
@item B\n\
An (row_idx(2)-row_idx(1)) x (col_idx(2)-col_idx(1)) matrix.\n\
@item n_cpus\n\
Number of threads to use. n_cpus must be greater or equal to 1.\n\
@end table\n\
\n\
Output parameter:\n\
\n\
@table @samp\n\
@item -\n\
There are no output arguments for @code{copy_p}.\n\
@end table\n\
\n\
@code{copy_p} is an oct-function that is a part of the DREAM Toolbox available at @url{http://www.signal.uu.se/Toolbox/dream/}.\n\
\n\
Copyright @copyright{} 2006-2008 Fredrik Lingvall.\n\
@end deftypefn")
{
  double         *A,*B; 
  sighandler_t   old_handler, old_handler_abrt, old_handler_keyint;
  pthread_t      *threads;
  dream_idx_type line_start, line_stop, A_M, A_N, B_M, B_N;
  int            thread_n,  N, err;    
  dream_idx_type r_M, r_N, k_M, k_N;
  double         *r, *k;
  void           *retval;
  DATA           *D;
  octave_value_list oct_retval; 

  int nrhs = args.length ();

  // Check for proper number of arguments

  if (nrhs != 5) {
    error("copy_p requires 5 input arguments!");
    return oct_retval;
  }

  if (nlhs > 0) {
    error("Too many output arguments for copy_p, none required!");
    return oct_retval;
  }
  
  const Matrix tmp = args(0).matrix_value();
  A_M = tmp.rows();
  A_N = tmp.cols();
  A = (double*) tmp.fortran_vec();

  const Matrix tmp2 = args(1).matrix_value();
  r_M = tmp2.rows();
  r_N = tmp2.cols();
  r = (double*) tmp2.fortran_vec();

  const Matrix tmp3 = args(2).matrix_value();
  k_M = tmp3.rows();
  k_N = tmp3.cols();
  k = (double*) tmp3.fortran_vec();

  const Matrix tmp4 = args(3).matrix_value();
  B_M = tmp4.rows();
  B_N = tmp4.cols();
  B = (double*) tmp4.fortran_vec();

  // Check that arg 2. 
  if ( r_M * r_N !=2 ) {
    error("Argument 2 must be a 2 element vector!");
    return oct_retval;
  }

  if ( (dream_idx_type) r[1] < 1) {
    error("1st element of argument 2 must be > 1!");
    return oct_retval;
  }

  if (( (dream_idx_type) r[0] > A_M) || ( (dream_idx_type) r[1] > A_M)) {
    error("One element of argument 2 exceeds row dimension of arg 1!");
    return oct_retval;
  }

  // Check that arg 3. 
  if ( k_M * k_N !=2 ) {
    error("Argument 3 must be a 2 element vector!");
    return oct_retval;
  }

  if ( (dream_idx_type) k[1] < 1) {
    error("1st element of argument 3 must be > 1!");
    return oct_retval;
  }
  
  if (( (dream_idx_type) k[0] > A_N) || ( (dream_idx_type) k[1] > A_N)) {
    error("One element of argument 3 exceeds colomn dimension of arg 1!");
    return oct_retval;
  }

  // Check that arg 4. 
  if (B_M < (dream_idx_type) (r[1]-r[0])+1 ) {
    error("Argument 4 has not the number of rows as given by arg 2!");
    return oct_retval;
  }

  if (B_N != (dream_idx_type) (k[1]-k[0])+1 ) {
    error("Argument 4 has not the number of columns as given by arg 3!");
    return oct_retval;
  }

    
  //
  // Number of threads.
  //

  // Check that arg 5 is a scalar.
  if (  args(4).matrix_value().rows()  * args(4).matrix_value().cols() !=1) {
    error("Argument 5 must be a scalar!");
    return oct_retval;
  }

  N = (dream_idx_type) args(4).matrix_value().fortran_vec()[0];
  if (N < 1) {
    error("Argument 5 must be larger or equal to 1 (min number of CPUs)!");
    return oct_retval;
  }

  // Check n_cpus argument.
  if (N > A_N) {
    // Add a -v verbose flag to display stuff like this!
    //mexPrintf("Warning: n_cpus is larger then number of columns in first arg.\n");
    //mexPrintf("         Setting n_cpus = # cols in 1st arg!\n");
    N = A_N; //
  }

  
  
  //
  // Register signal handlers.
  //

  if ((old_handler = signal(SIGTERM, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }

  if ((old_handler_abrt=signal(SIGABRT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  if ((old_handler_keyint=signal(SIGINT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  //
  // Call the mem_copy subroutine.
  //

  running = TRUE;
  
  // Allocate local data.
  D = (DATA*) malloc(N*sizeof(DATA));
  if (!D) {
    error("Failed to allocate memory for thread data!");
    return oct_retval;
  }

  // Allocate mem for the threads.
  threads = (pthread_t*) malloc(N*sizeof(pthread_t));
  
  if (!threads) {
    error("Failed to allocate memory for threads!");
    return oct_retval;
  }

  for (thread_n = 0; thread_n < N; thread_n++) {

    line_start = ((dream_idx_type) thread_n) * A_N/N;
    line_stop =  ((dream_idx_type) thread_n+1) * A_N/N;

    // Init local data.
    D[thread_n].line_start = line_start; // Local start index;
    D[thread_n].line_stop = line_stop; // Local stop index;
    D[thread_n].A = A;
    D[thread_n].A_M = A_M;
    D[thread_n].A_N = A_N;
    D[thread_n].B = B;
    D[thread_n].B_M = B_M;
    D[thread_n].B_N = B_N;
    D[thread_n].r = (dream_idx_type) (r[0]-1);
    D[thread_n].k = (dream_idx_type) (k[0]-1);
    D[thread_n].len = (dream_idx_type) (r[1]-r[0]+1);

    // Starts the threads.
    err = pthread_create(&threads[thread_n], NULL, smp_process, &D[thread_n]);
    if (err != 0) {
      error("Error when creating a new thread!\n");
      return oct_retval;
    }
  }
  
  // Wait for all threads to finish.
  for (thread_n = 0; thread_n < N; thread_n++) {
    err = pthread_join(threads[thread_n], &retval);
    if (err != 0) {
      error("Error when joining a thread!\n");
      return oct_retval;
    }
  }


  // Free memory.
  if (D) {
    free((void*) D);
  }

  //
  // Restore old signal handlers.
  //

  if (signal(SIGTERM, old_handler) == SIG_ERR) {
    printf("Couldn't register old signal handler.\n");
  }
   
  if (signal(SIGABRT,  old_handler_abrt) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  if (signal(SIGINT, old_handler_keyint) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }

  if (!running) {
    error("CTRL-C pressed!\n"); // Bail out.
    return oct_retval;
  }
  
  if (threads) {
    free((void*) threads);
  }

  return oct_retval;
}
