/***
*
* Copyright (C) 2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include "mex.h"
#include "rect_sir.h"
#include "dream_error.h"

//
// Octave headers.
//

#include <octave/oct.h>

#include <octave/config.h>

#include <iostream>
using namespace std;

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

//
// Macros
//

#define mxGetM(N)   args(N).matrix_value().rows()
#define mxGetN(N)   args(N).matrix_value().cols()
#define mxIsChar(N) args(N).is_string()

/***
 *
 *  rect_sir - Octave (oct) gateway function for RECT_SIR.
 *
 ***/

DEFUN_DLD (rect_sir, args, nlhs,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {}  [Y] = rect_sir(Ro,geom_par,delay,s_par,m_par).\n \
\n\
RECT_SIR Computes the time-continous (analytic) spatial impulse response(s) for a\n\
rectangular transducer.\n\
\n\
Observation point(s) ([mm]):\n\
\n\
@table @code\n\
@item Ro\n\
An N x 3 matrix, Ro = [xo1 yo1 zo2; xo2 yo2 zo2; ... xoN yoN zoN]; where N is the number of observation points.\n\
@end table\n\
\n\
Geometrical parameters: geom_par = [a b];\n\
\n\
@table @code\n\
@item a\n\
@item b\n\
Dimensions of the transducer [mm].\n\
@end table\n\
\n\
Sampling parameters: s_par = [dt nt]; \n\
\n\
@table @code\n\
@item dt\n\
Temporal discretization period (= 1/sampling freq) [us].\n\
@item  nt\n\
Length of impulse response vector.\n\
@end table\n\
\n\
Start point of SIR:\n\
\n\
@table @code\n\
@item  delay\n\
Scalar delay for all observation points or a vector with individual delays for each observation point [us].\n\
@end table\n\
\n\
Material parameters: m_par = [v cp];\n\
\n\
@table @code\n\
@item v\n\
Normal velocity [m/s].\n\
@item cp\n\
Sound velocity [m/s].\n\
\n\
@end table\n\
\n\
rect_sir is an oct-function that is a part of the DREAM Toolbox available at\n\
@url{http://www.signal.uu.se/Toolbox/dream/}.\n\
\n\
Copyright @copyright{} 2008 Fredrik Lingvall.\n\
@seealso {dreamrect,circ_sir}\n\
@end deftypefn")
{
  double *ro,*geom_par,*s_par,*m_par;
  dream_idx_type nt,no,n;
  double xo, yo, zo, a, b, dt;
  double *delay,v,cp;     
  double *h;
  octave_value_list oct_retval;

  int nrhs = args.length ();

  // Check for proper number of arguments
    
  if (nrhs != 5) {
    dream_err_msg("rect_sir requires 5 input arguments!");
  }
  else
    if (nlhs > 1) {
      dream_err_msg("Too many output arguments for rect_sir!");
    }
  
  //
  // Observation point.
  //
  
  // Check that arg (number of observation points) x 3 matrix
  if (!mxGetN(0)==3) {
    error("Argument 1 must be a (number of observation points) x 3 matrix!");
    return oct_retval;
  }
  
  no = mxGetM(0); // Number of observation points.
  const Matrix tmp0 = args(0).matrix_value();
  ro = (double*) tmp0.fortran_vec();

  //
  // Transducer geometry
  //

  // Check that arg 2 is a scalar.
  if (!((mxGetM(1)==2 && mxGetN(1)==1) || (mxGetM(1)==1 && mxGetN(1)==2))) {
    error("Argument 2 must be a two element vector!");
    return oct_retval;  
  }
  
  const Matrix tmp1 = args(1).matrix_value();
  geom_par = (double*) tmp1.fortran_vec();
  a  = geom_par[0];		// x-dim of the transducer.
  b  = geom_par[1];		// y-dim of the transducer.

  //
  // Temporal and spatial sampling parameters.
  //
  
  // Check that arg 3 is a 2 element vector
  if (!((mxGetM(2)==2 && mxGetN(2)==1) || (mxGetM(2)==1 && mxGetN(2)==2))) {
    error("Argument 3 must be a vector of length 2!");
    return oct_retval;
  }

  const Matrix tmp2 = args(2).matrix_value();
  s_par = (double*) tmp2.fortran_vec();
  dt    = s_par[0]; // Temporal discretization size (= 1/sampling freq).
  nt    = (dream_idx_type) s_par[1];	// Length of SIR.

  //
  // Start point of impulse response vector ([us]).
  //

  // Check that arg 4 is a scalar (or vector).
  if ( (mxGetM(3) * mxGetN(3) !=1) && ((mxGetM(3) * mxGetN(3)) != no)) {
    error("Argument 4 must be a scalar or a vector with a length equal to the number of observation points!");
    return oct_retval;
  }
  
  const Matrix tmp3 = args(3).matrix_value();
  delay = (double*) tmp3.fortran_vec();
  
  //
  // Material parameters
  //
  
  // Check that arg 5 is a 2 element vector.
  if (!((mxGetM(4)==2 && mxGetN(4)==1) || (mxGetM(4)==1 && mxGetN(4)==2))) {
    error("Argument 5 must be a vector of length 2!");
    return oct_retval;
  }
  
  const Matrix tmp4 = args(4).matrix_value();
  m_par = (double*) tmp4.fortran_vec();
  v     = m_par[0]; // Normal velocity of transducer surface.
  cp    = m_par[1]; // Sound speed.

  
  //  
  // Create an output matrix for the impulse response(s).
  //
  
  Matrix h_mat(nt, no);
  h = h_mat.fortran_vec();
  
  // Call the analytic SIR subroutine.
  if (mxGetM(3) * mxGetN(3) == 1) {
    for (n=0; n<no; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      
      rect_sir(xo,yo,zo,a,b,dt,nt,delay[0],v,cp,&h[n*nt]);
    }
  } else {
    for (n=0; n<no; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      
      rect_sir(xo,yo,zo,a,b,dt,nt,delay[n],v,cp,&h[n*nt]);
    }
  }

  oct_retval.append(h_mat);
    
  return oct_retval;
}
      
