/***
*
* Copyright (C) 2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "dream_error.h"

//
// Octave headers.
//

#include <octave/oct.h>

#include <octave/config.h>

#include <iostream>
using namespace std;

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

//
// Macros
//

#define mxGetM(N)   args(N).matrix_value().rows()
#define mxGetN(N)   args(N).matrix_value().cols()
#define mxIsChar(N) args(N).is_string()

/***
 *
 * Time-domain (monostatic) Synthetic Aperture Focusing Technique (SAFT)
 *
 ***/

/***
 * 
 * Octave (oct) gateway function for SAFT.
 *
 ***/

DEFUN_DLD (saft, args, nlhs,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {}  [Y] = saft(B,To,delay,s_par,m_par,Ro,a).\n\
\n\
SAFT - The sythetic aperture focusing techique.\n\
\n\
Observation point(s) ([mm]):\n\
\n\
@table @code\n\
@item B\n\
An K x L matrix B-scan data matrix.\n\
@end table\n\
\n\
@table @code\n\
@item To\n\
An L x 3 matrix, Ro = [xo1 yo1 zo2; xo2 yo2 zo2; ... xoN yoN zoN]; where L is the number of transducer positions.\n\
@end table\n\
\n\
Start point of SIR:\n\
\n\
@table @code\n\
@item  delay\n\
Scalar delay for all observation points or a vector with individual delays for each observation point [us].\n\
@end table\n\
\n\
Sampling parameters: s_par = [dt]; \n\
\n\
@table @code\n\
@item dt\n\
Temporal discretization period (= 1/sampling freq) [us].\n\
@end table\n\
\n\
Material parameters: m_par = [v cp];\n\
\n\
@table @code\n\
@item cp\n\
Sound velocity [m/s].\n\
\n\
@table @code\n\
@item Ro\n\
An N x 3 matrix, Ro = [xo1 yo1 zo2; xo2 yo2 zo2; ... xoN yoN zoN]; where N is the number of observation points.\n\
@end table\n\
\n\
@table @code\n\
@item  a\n\
The synthetic aperture [mm].\n\
@end table\n\
\n\
saft is an oct-function that is a part of the DREAM Toolbox available at\n\
@url{http://www.signal.uu.se/Toolbox/dream/}.\n\
\n\
Copyright @copyright{} 2008 Fredrik Lingvall.\n\
@seealso {saft_p,das,das_arr}\n\
@end deftypefn")
{
  register int n, k_shift, l;
  int K, L, no, n_l;
  double *s_par, *m_par;
  double *B, *Bsaft, z_prim, cp , a, dt, d, *ro, xo, yo, zo, r_xy;
  double *r_trans, x_trans, y_trans, z_trans, *delay;
  double z, tmp;
  octave_value_list oct_retval;
  
  int nrhs = args.length ();

  // Check for proper number of arguments
  
  if (nrhs != 7) {
    error("saft requires 7 input arguments!");
    return oct_retval;

  }
  else
    if (nlhs > 1) {
      error("Too many output arguments for saft!");
      return oct_retval;
    }
  
  //
  // B-scan matrix.
  //

  K = mxGetM(0); // Number of temporal samples.
  L = mxGetN(0); // Number of transducer positions (A-scans).
  const Matrix tmp0 = args(0).matrix_value();
  B = (double*) tmp0.fortran_vec();

  //
  // Transducer position matrix.
  //

  // Check that arg 2 is a L x 3 is a matrix.
  if (!(mxGetM(1)==L && mxGetN(1)==3)) {
    error("Arg 2 must be a (number of transducer positions) x 3 matrix!");
    return oct_retval;
  }
  const Matrix tmp1 = args(1).matrix_value();
  r_trans = (double*) tmp1.fortran_vec();

  //
  // Start point of impulse response vector ([us]).
  //
  
  // Check that arg 3 is a scalar (or vector).
  if ( (mxGetM(2) * mxGetN(2) !=1) && ((mxGetM(2) * mxGetN(2)) != L)) {
    error("Argument 3 (delay(s)) must be a scalar or a vector with a length equal to the number of A-scans!");
    return oct_retval;
  }

  const Matrix tmp2 = args(2).matrix_value();
  delay = (double*) tmp2.fortran_vec();

  //
  // Sampling parameter.
  //

  // Check that arg 4 is a scalar.
  if (!((mxGetM(3)==1 && mxGetN(3)==1))) {
    error("Argument 4 (temporal sampling period) must be a scalar");
    return oct_retval;
  }

  const Matrix tmp3 = args(3).matrix_value();
  s_par = (double*) tmp3.fortran_vec();
  dt    = s_par[0]; // Temporal discretization size (= 1/sampling freq).

  //
  // Material parameter.
  //

  // Check that arg 5 is a scalar.
  if (!((mxGetM(4)==1 && mxGetN(4)==1))) {
    dream_err_msg("Argument 5 (sound speed) must be a scalar!");
    return oct_retval;
  }

  const Matrix tmp4 = args(4).matrix_value();
  m_par = (double*) tmp4.fortran_vec();
  cp    = m_par[0]; // Sound speed.

  //
  // Observation point matrix.
  //

  // Check that arg 6 (number of observation points) x 3 matrix.
  if (!mxGetN(5)==3) {
    error("Argument 6 must be a (number of observation points) x 3 matrix!");
    return oct_retval;
  }
  
  no = mxGetM(5); // Number of observation points.
  const Matrix tmp5 = args(5).matrix_value();
  ro = (double*) tmp5.fortran_vec();

  //
  // Synthetic Aperture
  //
 
  // Check that arg 7 is scalar.
  if ((mxGetM(6)!=1) && (mxGetN(6)!=1)) {
    error("Argument 7 must be a scalar!");
    return oct_retval;
  }
  
  const Matrix tmp6 = args(6).matrix_value();
  a = (double) tmp6.fortran_vec()[0];

  //
  // Create an output matrix for the procssed image.
  //

  Matrix h_mat(no,1);
  Bsaft = h_mat.fortran_vec();
  
  //
  // SAFT processing with linear interpolation.
  //

  // Loop over all observation points.
  for (n=0; n<no; n++) {
    
    // Current observation point.
    xo = ro[n]; 
    yo = ro[n+1*no]; 
    zo = ro[n+2*no]; 
    
    Bsaft[n] = 0.0;
    
    // Loop over elements in the aperture.
    n_l = 0;
    for (l=0; l<L; l++) {
      
      x_trans = r_trans[l];
      y_trans = r_trans[l+1*L];
      z_trans = r_trans[l+2*L];
	
      // Horizontal distance between transducer and observation point.
      r_xy = sqrt( (x_trans - xo)*(x_trans - xo) + (y_trans - yo)*(y_trans - yo) );

      // if r_xy is inside the synthetic aperture.
      if (r_xy <= a/2) {
	n_l ++;

	// Vertical distance between transducer and observation point.
	z = zo - z_trans;
	  
	z_prim = sqrt(z*z + r_xy*r_xy);
	  
	// Sampling index at depth z (double path).
	//k = rint(  z/cp/dt*2*1e3 - us[0]/dt );
	  
	// Focusing shift.
	//shift = rint( 2*(z_prim - z)/cp/dt * 1e3 );
	  
	// Rounding err.
	//d = 2*(z_prim - z)/cp/dt * 1e3 - ((double) shift);
	  
	// Shifted sample index.
	//k_shift = k + shift;
	  
	// Temp. storage to avoid recomputing.
	//tmp = z/cp/dt*2*1e3 - us[0]/dt + 2*(z_prim - z)/cp/dt * 1e3;

	if (mxGetM(2) * mxGetN(2) == 1) 
	  tmp = (2*z_prim/cp * 1e3 - 2*delay[0])/dt; 
	else
	  tmp = (2*z_prim/cp * 1e3 - 2*delay[n])/dt; 
	  
	// Better to round just one time!
	k_shift = (int) rint(tmp);

	// Rounding err.
	d =  tmp - ((double) k_shift);
	  
	// Linear interpolation.
	if ((k_shift+1 < K) && (k_shift-1 > 0) ) {
	  if (d >=0) {
	    Bsaft[n] += (1.0 - d) * B[k_shift     + l*K];
	    Bsaft[n] += d * B[(k_shift+1) + l*K];
	  }
	  else {
	    Bsaft[n] -= d * B[k_shift     + l*K];
	    Bsaft[n] += (1.0 + d) * B[(k_shift-1) + l*K];
	  }
	} // if
      } // SAFT loop. 
	
      //Bsaft[n] /= (double) n_l; // Normalize with # elements in aperture.
    }
  }

  oct_retval.append(h_mat);
  
  return oct_retval;
}

  
