/***
*
* Copyright (C) 2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>
#include "dream_error.h"

//
// Octave headers.
//

#include <octave/oct.h>

#include <octave/config.h>

#include <iostream>
using namespace std;

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

//
// Macros
//

#define mxGetM(N)   args(N).matrix_value().rows()
#define mxGetN(N)   args(N).matrix_value().cols()
#define mxIsChar(N) args(N).is_string()

#define SINGLE 0
#define MULTIPLE 1

/***
 *
 * (Monostatic) Synthetic Aperture Focusing Technique - Parallel version.
 *
 ***/

//
// Globals
//

int running;

//
// typedef:s
//

typedef struct
{
  int    no;
  int    start;
  int    stop;
  double *ro;
  double dt;
  int delay_method;
  double *delay;
  double cp;
  double a;
  int    K;
  int    L;
  double *r_trans;
  double *B;
  double *Bsaft;
} DATA;

typedef void (*sighandler_t)(int);

//
// Function prototypes.
//
void* smp_process(void *arg);
void sighandler(int signum);
void sig_abrt_handler(int signum);
void sig_keyint_handler(int signum);

/***
 *
 * Thread function. 
 *
 ***/

void* smp_process(void *arg)
{
  int n, n_l,k_shift, l;
  DATA D = *(DATA *)arg;
  double *Bsaft = D.Bsaft;
  double *B = D.B;
  double dt=D.dt;
  int    no=D.no, K=D.K, L=D.L;
  double *delay=D.delay, *ro=D.ro, cp=D.cp;
  double xo, yo, zo, r_xy;
  int    start=D.start, stop=D.stop;
  double *r_trans = D.r_trans, a = D.a;
  double x_trans, y_trans, z_trans, d, z, z_prim, tmp;

  if (D.delay_method == SINGLE) {
    for (n=start; n<stop; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      
      // The SAFT loop.
      for (l=0; l<L; l++) {
	
	x_trans = r_trans[l];
	y_trans = r_trans[l+1*L];
	z_trans = r_trans[l+2*L];
	
	// Horizontal distance between transducer and observation point.
	r_xy = sqrt( (x_trans - xo)*(x_trans - xo) + (y_trans - yo)*(y_trans - yo) );
	
	// if r_xy is inside the synthetic aperture.
	if (r_xy <= a/2) {
	  n_l ++;
	  
	  // Vertical distance between transducer and observation point.
	  z = zo - z_trans;
	  
	  z_prim = sqrt(z*z + r_xy*r_xy);
	  tmp = (2*z_prim/cp * 1e3 - 2*delay[0])/dt; 

	  // Better to round just one time!
	  k_shift = (int) rint(tmp);
	  
	  // Rounding err.
	  d =  tmp - ((double) k_shift);
	  
	  // Linear interpolation.
	  if ((k_shift+1 < K) && (k_shift-1 > 0) ) {
	    if (d >=0) {
	      Bsaft[n] += (1.0 - d) * B[k_shift     + l*K];
	      Bsaft[n] += d * B[(k_shift+1) + l*K];
	    }
	    else {
	      Bsaft[n] -= d * B[k_shift     + l*K];
	      Bsaft[n] += (1.0 + d) * B[(k_shift-1) + l*K];
	    }
	  } // if
	}
      }
      
      if (!running) {
	printf("Thread for observation points %d -> %d bailing out!\n",start+1,stop);
	return(NULL);
      }
      
    }

  } else { // MULTIPLE delays.

    for (n=start; n<stop; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      
      // The SAFT loop.
      for (l=0; l<L; l++) {
	
	x_trans = r_trans[l];
	y_trans = r_trans[l+1*L];
	z_trans = r_trans[l+2*L];
	
	// Horizontal distance between transducer and observation point.
	r_xy = sqrt( (x_trans - xo)*(x_trans - xo) + (y_trans - yo)*(y_trans - yo) );
	
	// if r_xy is inside the synthetic aperture.
	if (r_xy <= a/2) {
	  n_l ++;
	  
	  // Vertical distance between transducer and observation point.
	  z = zo - z_trans;
	  
	  z_prim = sqrt(z*z + r_xy*r_xy);
	  tmp = (2*z_prim/cp * 1e3 - 2*delay[n])/dt; 
	  
	  // Better to round just one time!
	  k_shift = (int) rint(tmp);
	  
	  // Rounding err.
	  d =  tmp - ((double) k_shift);
	  
	  // Linear interpolation.
	  if ((k_shift+1 < K) && (k_shift-1 > 0) ) {
	    if (d >=0) {
	      Bsaft[n] += (1.0 - d) * B[k_shift     + l*K];
	      Bsaft[n] += d * B[(k_shift+1) + l*K];
	    }
	    else {
	      Bsaft[n] -= d * B[k_shift     + l*K];
	      Bsaft[n] += (1.0 + d) * B[(k_shift-1) + l*K];
	    }
	  } // if
	}
      }
      
      if (!running) {
	printf("Thread for observation points %d -> %d bailing out!\n",start+1,stop);
	return(NULL);
      }
    }
  }

  return(NULL);
}


/***
 *
 * Signal handlers.
 *
 ***/

void sighandler(int signum) {
  //printf("Caught signal SIGTERM.\n");
  running = FALSE;
}

void sig_abrt_handler(int signum) {
  //printf("Caught signal SIGABRT.\n");
}

void sig_keyint_handler(int signum) {
  //printf("Caught signal SIGINT.\n");
}

/***
 * 
 * Octave (oct) gateway function for SAFT_P.
 *
 ***/

DEFUN_DLD (saft_p, args, nlhs,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {}  [Y] = saft_p(B,To,delay,s_par,m_par,Ro,a).\n\
\n\
SAFT_P - The sythetic aperture focusing techique.\n\
\n\
Observation point(s) ([mm]):\n\
\n\
@table @code\n\
@item B\n\
An K x L matrix B-scan data matrix.\n\
@end table\n\
\n\
@table @code\n\
@item To\n\
An L x 3 matrix, Ro = [xo1 yo1 zo2; xo2 yo2 zo2; ... xoN yoN zoN]; where L is the number of transducer positions.\n\
@end table\n\
\n\
Start point of SIR:\n\
\n\
@table @code\n\
@item  delay\n\
Scalar delay for all observation points or a vector with individual delays for each observation point [us].\n\
@end table\n\
\n\
Sampling parameters: s_par = [dt]; \n\
\n\
@table @code\n\
@item dt\n\
Temporal discretization period (= 1/sampling freq) [us].\n\
@end table\n\
\n\
Material parameters: m_par = [v cp];\n\
\n\
@table @code\n\
@item cp\n\
Sound velocity [m/s].\n\
\n\
@table @code\n\
@item Ro\n\
An N x 3 matrix, Ro = [xo1 yo1 zo2; xo2 yo2 zo2; ... xoN yoN zoN]; where N is the number of observation points.\n\
@end table\n\
\n\
@table @code\n\
@item  a\n\
The synthetic aperture [mm].\n\
@end table\n\
\n\
\n\
Parallel processing parameter:\n\
@table @code\n\
@item n_cpus\n\
Number of parallel processes (threads). Must be larger or equal to 1.\n\
@end table\n\
\n\
saft_p is an oct-function that is a part of the DREAM Toolbox available at\n\
@url{http://www.signal.uu.se/Toolbox/dream/}.\n\
\n\
Copyright @copyright{} 2008 Fredrik Lingvall.\n\
@seealso {saft,das,das_arr}\n\
@end deftypefn")
{
  double *s_par,*m_par;
  int K, L, no, err;
  double *B, *Bsaft, cp , a, dt, *ro;
  double *r_trans, *delay;
  sighandler_t old_handler, old_handler_abrt, old_handler_keyint;
  pthread_t *threads;
  int    thread_n, n_cpus=1, start,stop;
  void   *retval;
  DATA   *D;
  octave_value_list oct_retval;
  
  int nrhs = args.length ();

  // Check for proper number of arguments
  
  if (nrhs != 8) {
    error("saft_p requires 8 input arguments!");
    return oct_retval;
  }
  else
    if (nlhs > 1) {
      error("Too many output arguments for saft_p!");
      return oct_retval;
    }
  
  //
  // B-scan.
  //

  K = mxGetM(0); // Number of temporal samples.
  L = mxGetN(0); // Number of transducer positions.
  const Matrix tmp0 = args(0).matrix_value();
  B = (double*) tmp0.fortran_vec();

  //
  // Transducer position matrix.
  //

  // Check that arg 2 is a L x 3 is a matrix.
  if (!(mxGetM(1)==L && mxGetN(1)==3)) {
    error("Arg 2 must be a (number of transducer positions) x 3 matrix!");
    return oct_retval;
  }
  const Matrix tmp1 = args(1).matrix_value();
  r_trans = (double*) tmp1.fortran_vec();

  //
  // Start point of impulse response vector ([us]).
  //
  
  // Check that arg 3 is a scalar (or vector).
  if ( (mxGetM(2) * mxGetN(2) !=1) && ((mxGetM(2) * mxGetN(2)) != L)) {
    error("Argument 3 (delay(s)) must be a scalar or a vector with a length equal to the number of A-scans!");
    return oct_retval;
  }

  const Matrix tmp2 = args(2).matrix_value();
  delay = (double*) tmp2.fortran_vec();

  //
  // Sampling parameter.
  //

  // Check that arg 4 is a scalar.
  if (!((mxGetM(3)==1 && mxGetN(3)==1))) {
    error("Argument 4 (temporal sampling period) must be a scalar");
    return oct_retval;
  }

  const Matrix tmp3 = args(3).matrix_value();
  s_par = (double*) tmp3.fortran_vec();
  dt    = s_par[0]; // Temporal discretization size (= 1/sampling freq).

  //
  // Material parameter.
  //

  // Check that arg 5 is a scalar.
  if (!((mxGetM(4)==1 && mxGetN(4)==1))) {
    dream_err_msg("Argument 5 (sound speed) must be a scalar!");
    return oct_retval;
  }

  const Matrix tmp4 = args(4).matrix_value();
  m_par = (double*) tmp4.fortran_vec();
  cp    = m_par[0]; // Sound speed.
  
  //
  // Observation point matrix.
  //

  // Check that arg 6 (number of observation points) x 3 matrix.
  if (!mxGetN(5)==3) {
    error("Argument 6 must be a (number of observation points) x 3 matrix!");
    return oct_retval;
  }
  
  no = mxGetM(5); // Number of observation points.
  const Matrix tmp5 = args(5).matrix_value();
  ro = (double*) tmp5.fortran_vec();

  //
  // Synthetic Aperture
  //
 
  // Check that arg 7 is scalar.
  if ((mxGetM(6)!=1) && (mxGetN(6)!=1)) {
    error("Argument 7 must be a scalar!");
    return oct_retval;
  }
  
  const Matrix tmp6 = args(6).matrix_value();
  a = (double) tmp6.fortran_vec()[0];

  //
  // Number of threads.
  //

  // Check that arg 8 is a scalar.
  if ( mxGetM(7) * mxGetN(7) !=1) {
    error("Argument 8 must be a scalar!");
    return oct_retval; 
  }
  const Matrix tmp7 = args(7).matrix_value();
  n_cpus = (int) tmp7.fortran_vec()[0];
  if (n_cpus < 1) {
    error("Argument 7 must be larger or equal to 1 (min number of CPUs)!");
    return oct_retval; 
  }

  // n_cpus can't be larger then the number of observation points.
  if (n_cpus > no) { 
    n_cpus = no;
  }
  
  //
  // Create an output matrix for the processed image.
  //

  Matrix h_mat(no,1);
  Bsaft = h_mat.fortran_vec();

  //
  // Register signal handlers.
  //

  if ((old_handler = signal(SIGTERM, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }

  if ((old_handler_abrt=signal(SIGABRT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  if ((old_handler_keyint=signal(SIGINT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  //
  // Call the SAFT subroutine.
  //

  running = TRUE;

  // Allocate local data.
  D = (DATA*) malloc(n_cpus*sizeof(DATA));

  // Allocate mem for the threads.
  threads = (pthread_t*) malloc(n_cpus*sizeof(pthread_t));
  
  for (thread_n = 0; thread_n < n_cpus; thread_n++) {

    start = thread_n * no/n_cpus;
    stop =  (thread_n+1) * no/n_cpus;

    // Init local data.
    D[thread_n].start = start; // Local start index;
    D[thread_n].stop = stop; // Local stop index;
    D[thread_n].no = no;
    D[thread_n].ro = ro;
    D[thread_n].cp = cp;
    D[thread_n].dt = dt;

    if (mxGetM(3) * mxGetN(3) == 1)
      D[thread_n].delay_method = SINGLE; // delay is a scalar.
    else
      D[thread_n].delay_method = MULTIPLE; // delay is a vector.

    D[thread_n].delay = delay;
    D[thread_n].B = B;
    D[thread_n].Bsaft = Bsaft;
    D[thread_n].a = a;
    D[thread_n].K = K;
    D[thread_n].L = L;
    D[thread_n].r_trans = r_trans;

    // Starts the threads.
    err = pthread_create(&threads[thread_n], NULL, smp_process, &D[thread_n]);
    if (err != 0) {
      error("Error when creating a new thread!\n");
      return oct_retval;
    }
  }
  
  // Wait for all threads to finish.
  for (thread_n = 0; thread_n < n_cpus; thread_n++) {
    err = pthread_join(threads[thread_n], &retval);
    if (err != 0) {
      error("Error when joining a thread!\n");
      return oct_retval;
    }
  }

  // Free memory.
  free((void*) D);

  //
  // Restore old signal handlers.
  //

  if (signal(SIGTERM, old_handler) == SIG_ERR) {
    printf("Couldn't register old signal handler.\n");
  }
   
  if (signal(SIGABRT,  old_handler_abrt) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }
  
  if (signal(SIGINT, old_handler_keyint) == SIG_ERR) {
    printf("Couldn't register signal handler.\n");
  }

  if (!running) {
    error("CTRL-C pressed!\n"); // Bail out.
    return oct_retval;
  }

  oct_retval.append(h_mat);

  return oct_retval;
}

