/***
*
* Copyright (C) 2002,2003,2006,2007,2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <math.h>
#include <stdlib.h>

#include "mex.h"
#include "fft.h"

#ifdef USE_FFTW
#include <fftw3.h>
#endif


/*** FFTW real-to-real transform.
 *
 * FFTW_R2HC computes a real-input DFT with output in \u201chalfcomplex\u201d format, i.e.
 *  real and imaginary parts for a transform of size n stored as:
 *
 * r0, r1, r2, ..., rn/2, i(n+1)/2-1, ..., i2, i1
 * (Logical N=n, inverse is FFTW_HC2R.)
 *
 * FFTW_HC2R computes the reverse of FFTW_R2HC, above. (Logical N=n, inverse is FFTW_R2HC.) 
 *
 ***/

#ifdef USE_FFTW

// Global plans.
fftw_plan    p_forward;
fftw_plan    p_backward;

void fft_init(dream_idx_type n, fftw_complex *xc, double *y)
{
  p_backward = fftw_plan_dft_c2r_1d(n, xc, y, FFTW_MEASURE);

  return;
}

void fft_close()
{
  fftw_destroy_plan(p_backward); 

  return;
}

#endif

/***
 *
 * Amplitude spectrum of a real vector. (not used in DREAM)
 *
 ***/

void dream_fft(double *x, double *y, dream_idx_type n) // Should be renamed to rr_fft.
{
  dream_idx_type i;

#ifdef USE_FFTW


  // Fourier transform xr.
  fftw_execute_r2r(p_forward,x,y);

  // Compute absolute value.
  for(i=0; i<n/2; i++)
    y[i] = sqrt(y[i]*y[i] + y[n-i]*y[n-i]);
  for(i=n/2; i<n; i++)
    y[i] = y[n-i];

#else

  mxArray *X,*Y;
  double *yr,*yi,*xr;

  X = mxCreateDoubleMatrix(n,1,mxREAL);
  xr = mxGetPr(X);
  
  // Copy input vector to a Matlab Matrix.
  for (i=0; i<n; i++)
    xr[i] = x[i];
  
  // Let Matlab do the job!
  mexCallMATLAB(1, &Y, 1, &X, "fft");
  yr = mxGetPr(Y);
  yi = mxGetPi(Y);

  for (i=0; i<n; i++)
    y[i] = sqrt(yr[i]*yr[i] + yi[i]*yi[i]);
  
  mxDestroyArray(X);
  mxDestroyArray(Y);

#endif

  return;
}

/***
 *
 * Inverse DFT of a real vector (returns the real part).  (not used in DREAM)
 *
 ***/

void dream_ifft(double *x, double *y, dream_idx_type n) // Should be renamed to rr_ifft.
{
  dream_idx_type       i;

#ifdef USE_FFTW

  fftw_plan p_backward;


  // Set imag part to zero.
  for(i=n/2; i<n; i++)
    x[i] = 0.0;

  // Create real-to-real FFTW plan.
  p_backward = fftw_plan_r2r_1d(n, x, y, FFTW_HC2R, FFTW_MEASURE);

  // Fourier transform xr.
  fftw_execute_r2r(p_backward,x,y);

#else

  mxArray *X,*Y;
  double *yr,*yi,*xr;

  X = mxCreateDoubleMatrix(n,1,mxREAL);
  xr = mxGetPr(X);
  
  // Copy input vector to a Matlab Matrix.
  for (i=0; i<n; i++)
    xr[i] = x[i];
  
  // Let Matlab do the job!
  mexCallMATLAB(1, &Y, 1, &X, "ifft");
  yr = mxGetPr(Y);
  yi = mxGetPi(Y);

  for (i=0; i<n; i++)
    y[i] = yr[i];
  
  mxDestroyArray(X);
  mxDestroyArray(Y);

#endif

  return;
}

/***
 *
 * Inverse DFT of a complex vector (returns the real part).
 *
 ***/

#ifdef USE_FFTW
void cr_ifft(fftw_complex *xc, double *y, dream_idx_type n)
#else
void cr_ifft(double *xir, double *xii, double *y, dream_idx_type n)
#endif
{

#ifdef USE_FFTW
  dream_idx_type k;

  fftw_execute_dft_c2r(p_backward,xc,y);
  
  // Normalize.
  for (k=0; k<n; k++)
    y[k] /= (double) n;

#else

  mxArray *X,*Y;
  double *yr,*yi,*xr,*xi;
  dream_idx_type k;

  X = mxCreateDoubleMatrix(n,1,mxCOMPLEX);
  xr = mxGetPr(X);
  xi = mxGetPi(X);
  
  // Copy input vector to a Matlab Matrix.
  for (k=0; k<n; k++) {
    xr[k] = xir[k];
    xi[k] = xii[k];
  }
  
  // Let Matlab do the job!
  mexCallMATLAB(1, &Y, 1, &X, "ifft");
  yr = mxGetPr(Y);
  yi = mxGetPi(Y);

  // Use the real part only.
  for (k=0; k<n; k++)
    y[k] = yr[k];

  mxDestroyArray(X);
  mxDestroyArray(Y);

#endif
  return;
}


/***
 *
 * Inverse DFT of a complex vector. (not used in DREAM)
 *
 ***/

#ifdef USE_FFTW
void cc_ifft(fftw_complex *xc, fftw_complex *yc, dream_idx_type n)
#else
void cc_ifft(double *xir, double *xii, double *yor, double *yoi, dream_idx_type n)
#endif
{

#ifdef USE_FFTW

  dream_idx_type k;

  fftw_execute_dft(p_backward,xc,yc);
  
  // Normalize.
  for (k=0; k<n; k++)
    yc[k] /= (double) n;

#else
  dream_idx_type i;
  mxArray *X,*Y;
  double *yr,*yi,*xr,*xi;

  X = mxCreateDoubleMatrix(n,1,mxCOMPLEX);
  xr = mxGetPr(X);
  xi = mxGetPi(X);
  
  // Copy input vector to a Matlab Matrix.
  for (i=0; i<n; i++) {
    xr[i] = xir[i];
    xi[i] = xii[i];
  }

  // Let Matlab do the job!
  mexCallMATLAB(1, &Y, 1, &X, "ifft");
  yr = mxGetPr(Y);
  yi = mxGetPi(Y);
  
  // Copy data.
  for (i=0; i<n; i++) {
    yor[i] = yr[i];
    yoi[i] = yi[i];
  }
  
  mxDestroyArray(X);
  mxDestroyArray(Y);

#endif

  return;
}
