/***
*
* Copyright (C) 2003,2005,2006,2007,2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <signal.h>
#include "mex.h"
#include "dreamcirc.h"
#include "dream_error.h"

#define SINGLE 0
#define MULTIPLE 1

#ifdef USE_FFTW
#include "att.h"
#endif

// Parallel implementation using Linux threads.

//
// Globals
//
volatile int out_err = NONE;
pthread_mutex_t err_lock; 
int running;

//
// typedef:s
//

typedef struct
{
  dream_idx_type no;
  dream_idx_type start;
  dream_idx_type stop;
  double *ro;
  double r; 
  double dx;
  double dy;
  double dt; 
  dream_idx_type nt;
  int delay_method;
  double *delay;
  double v;
  double cp;
  double alfa;
  double *h; 
  int err_level;
} DATA;

typedef void (*sighandler_t)(int);

//
// Function prototypes.
//
void* smp_process(void *arg);
void sighandler(int signum);
void sig_abrt_handler(int signum);
void sig_keyint_handler(int signum);

/***
 *
 * Thread function. 
 *
 ***/

void* smp_process(void *arg)
{
  int tmp_err = NONE, err = NONE;
  DATA D = *(DATA *)arg;
  double xo, yo, zo;
  double *h = D.h;
  double r=D.r, dx=D.dx, dy=D.dy, dt=D.dt;
  dream_idx_type n, no=D.no, nt=D.nt;
  int    tmp_lev, err_level=D.err_level;
  double *delay=D.delay, *ro=D.ro, v=D.v, cp=D.cp, alfa=D.alfa;
  dream_idx_type start=D.start, stop=D.stop;

  // Let the thread finish and then catch the error. 
  if (err_level == STOP)
    tmp_lev = PARALLEL_STOP;
  else
    tmp_lev = err_level;

  if (D.delay_method == SINGLE) {
    for (n=start; n<stop; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      err = dreamcirc(xo,yo,zo,r,dx,dy,dt,nt,delay[0],v,cp,alfa,
		      &h[n*nt],tmp_lev); 

      if (err != NONE || out_err ==  PARALLEL_STOP) {
	tmp_err = err;
	if (err == PARALLEL_STOP || out_err ==  PARALLEL_STOP)
	  break; // Jump out when a STOP error occurs.
      }
      
      if (!running) {
	printf("Thread for observation points %d -> %d bailing out!\n",start+1,stop);
	return(NULL);
      }

    }
  } else { // MULTIPLE delays.
    for (n=start; n<stop; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      err = dreamcirc(xo,yo,zo,r,dx,dy,dt,nt,delay[n],v,cp,alfa,
		      &h[n*nt],tmp_lev); 

      if (err != NONE || out_err ==  PARALLEL_STOP) {
	tmp_err = err;
	if (err == PARALLEL_STOP || out_err ==  PARALLEL_STOP)
	  break; // Jump out when a STOP error occurs.
      }
      
      if (!running) {
	printf("Thread for observation points %d -> %d bailing out!\n",start+1,stop);
	return(NULL);
      }

    }
  }

  // Lock out_err for update, update it, and unlock. 
  pthread_mutex_lock(&err_lock);
  
  if ((tmp_err != NONE) && (out_err == NONE))
    out_err = tmp_err;

  pthread_mutex_unlock(&err_lock);

  return(NULL);
}

/***
 *
 * Signal handlers.
 *
 ***/

void sighandler(int signum) {
  //printf("Caught signal SIGTERM.\n");
  running = FALSE;
}

void sig_abrt_handler(int signum) {
  //printf("Caught signal SIGABRT.\n");
}

void sig_keyint_handler(int signum) {
  //printf("Caught signal SIGINT.\n");
}

/***
 *
 * Matlab (MEX) gateway function for parallel dreamcirc.
 *
 ***/

void  mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  double *ro,*geom_par,*s_par,*m_par;
  dream_idx_type    nt,no;
  double r, dx, dy, dt;
  double *delay,v,cp,alfa;     
  double *h, *err_p;
  int    err_level=STOP, set = FALSE, err;
  char   err_str[50];
  int    buflen;
  DATA   *D;
  dream_idx_type    start, stop;
  pthread_t *threads;
  int    thread_n, N;
  void   *retval;
  sighandler_t   old_handler, old_handler_abrt, old_handler_keyint;

  // Check for proper number of arguments
    
  if (!((nrhs == 6) || (nrhs == 7))) {
    dream_err_msg("dreamcirc_p requires 6 or 7 input arguments!");
  }
  else
    if (nlhs > 2) {
      dream_err_msg("Too many output arguments for dreamcirc_p!");
    }
  
  //
  // Observation point.
  //
  
 // Check that arg (number of observation points) x 3 matrix
  if (!mxGetN(prhs[0])==3)
    dream_err_msg("Argument 1 must be a (number of observation points) x 3 matrix!");
  
  no = mxGetM(prhs[0]); // Number of observation points.
  ro = mxGetPr(prhs[0]);

  //if (no<2)
  //  dream_err_msg("At least 2 observation points i needed for this function!\n Use the serial version for a single observation point.");

  //
  // Transducer geometry
  //

   // Check that arg 2 is a scalar.
  if (!((mxGetM(prhs[1])==1 && mxGetN(prhs[1])==1)))
    dream_err_msg("Argument 2 must be a scalar!");
  
  geom_par = mxGetPr(prhs[1]);
  //xs = geom_par[0];		// x-pos (center point).
  //ys = geom_par[1];		// y-pos (center point).
  //a  = geom_par[2];		// Radius of the transducer.
  //xs = 0;			// x-pos (center point).
  //ys = 0;			// y-pos (center point).
  r  = geom_par[0];		// Radius of the transducer.

  //
  // Temporal and spatial sampling parameters.
  //
  
  // Check that arg 3 is a 4 element vector
  if (!((mxGetM(prhs[2])==4 && mxGetN(prhs[2])==1) || (mxGetM(prhs[2])==1 && mxGetN(prhs[2])==4)))
    dream_err_msg("Argument 3 must be a vector of length 4!");
  
  s_par = mxGetPr(prhs[2]);
  dx    = s_par[0];		// Spatial x discretization size.
  dy    = s_par[1];		// Spatial dy iscretization size.
  dt    = s_par[2];		// Temporal discretization size (= 1/sampling freq).
  nt    = (dream_idx_type) s_par[3];	// Length of SIR.
  
  //
  // Start point of impulse response vector ([us]).
  //

  // Check that arg 4 is a scalar.
  if ( (mxGetM(prhs[3]) * mxGetN(prhs[3]) !=1) && ((mxGetM(prhs[3]) * mxGetN(prhs[3])) != no))
    dream_err_msg("Argument 4 must be a scalar or a vector with a length equal to the number of observation points!");
  
  delay = mxGetPr(prhs[3]);
  
  //
  // Material parameters
  //
  
  // Check that arg 5 is a 3 element vectora
  if (!((mxGetM(prhs[4])==3 && mxGetN(prhs[4])==1) || (mxGetM(prhs[4])==1 && mxGetN(prhs[4])==3)))
    dream_err_msg("Argument 5 must be a vector of length 3!");
  
  m_par = mxGetPr(prhs[4]);
  v     = m_par[0]; // Normal velocity of transducer surface.
  cp    = m_par[1]; // Sound speed.
  alfa  = m_par[2]; // Attenuation coefficient [dB/(cm MHz)].
  
  //
  // Number of threads.
  //

  // Check that arg 6 is a scalar.
  if ( mxGetM(prhs[5]) * mxGetN(prhs[5]) !=1)
    dream_err_msg("Argument 5 must be a scalar!");
  
  N = (int) mxGetScalar(prhs[5]);
  if (N < 1)
    dream_err_msg("Argument 6 must be larger or equal to 1 (min number of CPUs)!");

  // n_cpus can't be larger then the number of observation points.
  if (N > no) { 
    N = no;
  }

  //  
  // Error reporting.
  //
  if (nrhs == 7) { // Add check a that arg is a string!
    
    if (!mxIsChar(prhs[6]))
      dream_err_msg("Argument 7 must be a string");  
    
    buflen = (mxGetM(prhs[6]) * mxGetN(prhs[6]) * sizeof(mxChar)) + 1;
    mxGetString(prhs[6],err_str,buflen);
    
    if (!strcmp(err_str,"ignore")) {
      err_level = IGNORE; 
      set = TRUE;
    }

    if (!strcmp(err_str,"warn")) {
      err_level = WARN; 
      set = TRUE;
    }

    if (!strcmp(err_str,"stop")) {
      err_level = STOP; 
      set = TRUE;
    }

    if (set == FALSE)
      dream_err_msg("Unknown error level!");
    
  }
  else
    err_level = STOP; // Default.
  
  
  // Create an output matrix for the impulse response
  plhs[0] = mxCreateDoubleMatrix(nt,no,mxREAL);
  h = mxGetPr(plhs[0]);

  //
  // Register signal handlers.
  //

  if ((old_handler = signal(SIGTERM, &sighandler)) == SIG_ERR) {
    printf("Couldn't register SIGTERM signal handler.\n");
  }

  if (( old_handler_abrt=signal(SIGABRT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register SIGABRT signal handler.\n");
  }
  
  if (( old_handler_keyint=signal(SIGINT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register SIGINT signal handler.\n");
  }
 
  //
  // Call the DREAM subroutine.
  //
  
  out_err = NONE;
  running=TRUE;

#ifdef USE_FFTW
  if (alfa != (double) 0.0)
    att_init(nt,N);
#endif

  // Initialize the lock on out_err.
  pthread_mutex_init(&err_lock, NULL);

  // Allocate local data.
  D = (DATA*) malloc(N*sizeof(DATA));

  // Allocate mem for the threads.
  threads = (pthread_t*) malloc(N*sizeof(pthread_t));
  
  for (thread_n = 0; thread_n < N; thread_n++) {

    start = thread_n * no/N;
    stop =  (thread_n+1) * no/N;

    // Init local data.
    D[thread_n].start = start; // Local start index;
    D[thread_n].stop = stop; // Local stop index;
    D[thread_n].no = no;
    D[thread_n].ro = ro;
    D[thread_n].r = r;
    D[thread_n].dx = dx;
    D[thread_n].dy = dy;
    D[thread_n].dt = dt; 
    D[thread_n].nt = nt;

    if (mxGetM(prhs[3]) * mxGetN(prhs[3]) == 1)
      D[thread_n].delay_method = SINGLE; // delay is a scalar.
    else
      D[thread_n].delay_method = MULTIPLE; // delay is a vector.

    D[thread_n].delay = delay;
    D[thread_n].v = v;
    D[thread_n].cp = cp;
    D[thread_n].alfa = alfa;
    D[thread_n].h = h;
    D[thread_n].err_level = err_level;
    
    // Starts the threads.
    err = pthread_create(&threads[thread_n], NULL, smp_process, &D[thread_n]);
    if (err != 0)
      dream_err_msg("Error when creating a new thread!\n");
  }
  
  // Wait for all threads to finish.
  for (thread_n = 0; thread_n < N; thread_n++) {
    err = pthread_join(threads[thread_n], &retval);
    if (err != 0)
      dream_err_msg("Error when joining a thread!\n");
  }

  // Free memory.
  free((void*) D);

  //
  // Restore old signal handlers.
  //

  if (signal(SIGTERM, old_handler) == SIG_ERR) {
    printf("Couldn't register old SIGTERM signal handler.\n");
  }
   
  if (signal(SIGABRT,  old_handler_abrt) == SIG_ERR) {
    printf("Couldn't register old SIGABRT signal handler.\n");
  }
  
  if (signal(SIGINT, old_handler_keyint) == SIG_ERR) {
    printf("Couldn't register old SIGINT signal handler.\n");
  }

#ifdef USE_FFTW
  if (alfa != (double) 0.0)
    att_close();
#endif 

  if (!running) {
    dream_err_msg("CTRL-C pressed!\n"); // Bail out.
  }

  //
  // Check for Error.
  //

  if ( (err_level == STOP) && (out_err != NONE))
    dream_err_msg(""); // Bail out if error.
  
  //  
  // Return error.
  //
  if (nlhs == 2) {
    plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);
    err_p =  mxGetPr(plhs[1]);
    err_p[0] = (double) out_err;
  }
  
  return;
}
      
