/***
*
* Copyright (C) 2006,2007,2008,2009 Fredrik Lingvall
*
* This file is part of the DREAM Toolbox.
*
* The DREAM Toolbox is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* The DREAM Toolbox is distributed in the hope that it will be useful, but 
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License
* along with the DREAM Toolbox; see the file COPYING.  If not, write to the 
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA.
*
***/

// $Revision: 565 $ $Date: 2009-09-17 22:24:06 +0200 (Thu, 17 Sep 2009) $ $LastChangedBy: dream $

#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <signal.h>
#include "dreamrect_f.h"
#include "dream_error.h"

#define SINGLE 0
#define MULTIPLE 1

#ifdef USE_FFTW
#include "att.h"
#endif

//
// Octave headers.
//

#include <octave/oct.h>

#include <octave/config.h>

#include <iostream>
using namespace std;

#include <octave/defun-dld.h>
#include <octave/error.h>
#include <octave/oct-obj.h>
#include <octave/pager.h>
#include <octave/symtab.h>
#include <octave/variables.h>

//
// Macros
//

#define mxGetM(N)   args(N).matrix_value().rows()
#define mxGetN(N)   args(N).matrix_value().cols()
#define mxIsChar(N) args(N).is_string()

//
// Globals
//

volatile int out_err = NONE;
pthread_mutex_t err_lock; 
int running;

//
// typedef:s
//

typedef struct
{
  dream_idx_type no;
  dream_idx_type start;
  dream_idx_type stop;
  double *ro;
  double a; 
  double b; 
  double dx;
  double dy;
  double dt; 
  dream_idx_type nt;
  int delay_method;
  double *delay;
  double v;
  double cp;
  double alfa;
  int ifoc;
  double focal;
  double *h; 
  int err_level;
} DATA;

typedef void (*sighandler_t)(int);

//
// Function prototypes.
//
void* smp_process(void *arg);
void sighandler(int signum);
void sig_abrt_handler(int signum);
void sig_keyint_handler(int signum);

/***
 *
 * Thread function. 
 *
 ***/
void* smp_process(void *arg)
{
  int tmp_err = NONE, err = NONE;
  DATA D = *(DATA *)arg;
  double xo, yo, zo;
  double *h = D.h;
  double a=D.a, b=D.b, dx=D.dx, dy=D.dy, dt=D.dt;
  dream_idx_type n, no=D.no, nt=D.nt;
  int    tmp_lev, err_level=D.err_level;
  double *delay=D.delay, *ro=D.ro, v=D.v, cp=D.cp, alfa=D.alfa, focal=D.focal;
  dream_idx_type start=D.start, stop=D.stop;
  int    ifoc = D.ifoc;

  // Let the thread finish and then catch the error. 
  if (err_level == STOP)
    tmp_lev = PARALLEL_STOP;
  else
    tmp_lev = err_level;

  if (D.delay_method == SINGLE) {
    for (n=start; n<stop; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      err = dreamrect_f(xo,yo,zo,a,b,dx,dy,dt,nt,delay[0],v,cp,alfa,
		      ifoc,focal,&h[n*nt],tmp_lev); 

      if (err != NONE || out_err ==  PARALLEL_STOP) {
	tmp_err = err;
	if (err == PARALLEL_STOP || out_err ==  PARALLEL_STOP)
	  break; // Jump out when a STOP error occurs.
      }
      
      if (!running) {
	printf("Thread for observation points %d -> %d bailing out!\n",start+1,stop);
	return(NULL);
      }

    }
  } else { // MULTIPLE delays.
    for (n=start; n<stop; n++) {
      xo = ro[n]; 
      yo = ro[n+1*no]; 
      zo = ro[n+2*no]; 
      err = dreamrect_f(xo,yo,zo,a,b,dx,dy,dt,nt,delay[n],v,cp,alfa,
		      ifoc,focal,&h[n*nt],tmp_lev); 

      if (err != NONE || out_err ==  PARALLEL_STOP) {
	tmp_err = err;
	if (err == PARALLEL_STOP || out_err ==  PARALLEL_STOP)
	  break; // Jump out when a STOP error occurs.
      }

      if (!running) {
	printf("Thread for observation points %d -> %d bailing out!\n",start+1,stop);
	return(NULL);
      }

    }
  }

  // Lock out_err for update, update it, and unlock. 
  pthread_mutex_lock(&err_lock);
  
  if ((tmp_err != NONE) && (out_err == NONE))
    out_err = tmp_err;

  pthread_mutex_unlock(&err_lock);

  return(NULL);
}


/***
 *
 * Signal handlers.
 *
 ***/

void sighandler(int signum) {
  //printf("Caught signal SIGTERM.\n");
  running = FALSE;
}

void sig_abrt_handler(int signum) {
  //printf("Caught signal SIGABRT.\n");
}

void sig_keyint_handler(int signum) {
  //printf("Caught signal SIGINT.\n");
}


/***
 *
 *  oct_dreamrect_f_pg.cc - Octave (oct) gateway function for parallel dreamrect_f.
 *
 ***/
DEFUN_DLD (dreamrect_f_p, args, nlhs,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {} [H,err] = dreamrect_f_p(Ro,geom_par,s_par,delay,m_par,foc_met,focal,n_cpus,err_level)\n\
\n\
DREAMRECT_F_P - Computes the spatial impulse response\n\
for a rectangular focused transducer using parallel processing \n\
(using threads).\n\
\n\
Observation point(s) ([mm]):\n\
\n\
@table @code\n\
@item Ro\n\
An N x 3 matrix, Ro = [xo1 yo1 zo2; xo2 yo2 zo2; ... xoN yoN zoN]; where N is the number of observation points.\n\
@end table\n\
\n\
Geometrical parameters: geom_par = [a b];\n\
\n\
@table @code\n\
@item a\n\
x-size  of the transducer [mm].\n\
@item b\n\
y-size  of the transducer [mm].\n\
@end table\n\
\n\
Sampling parameters: s_par = [dx dy dt nt]; \n\
\n\
@table @code\n\
@item dx\n\
Spatial x-direction discretization size [mm].\n\
@item dy\n\
Spatial y-direction discretization size [mm].\n\
@item dt\n\
Temporal discretization period (= 1/sampling freq) [us].\n\
@item  nt\n\
Length of impulse response vector.\n\
@end table\n\
\n\
@table @code\n\
@item  delay\n\
Scalar delay for all observation points or a vector with individual delays for each observation point [us].\n\
@end table\n\
\n\
Material parameters: m_par = [v cp alfa];\n\
\n\
@table @code\n\
@item v\n\
Normal velocity [m/s].\n\
@item cp\n\
Sound velocity [m/s].\n\
@item alfa\n\
Attenuation coefficient [dB/(cm MHz)].\n\
\n\
@end table\n\
\n\
Focusing parameters: foc_met and focal:\n\
\n\
@table @code\n\
@item foc_met\n\
Focusing method, options are: 'off', 'x', 'y', 'xy', and 'x+y'.\n\
@item  focal\n\
Focal distance [mm].\n\
@end table\n\
\n\
Parallel processing parameter:\n\
@table @code\n\
@item n_cpus\n\
Number of parallel processes (threads). Must be larger or equal to 1.\n\
@end table\n\
\n\
Error Handling: err_level;\n\
err_level is an optional text string parameter for controlling the error behavior, options are:\n\
\n\
@table @code\n\
@item 'ignore'\n\
An error is ignored (no error message is printed and the program is not stopped) but the err output \n\
argument is negative if an error occured.\n\
@item 'warn'\n\
An error message is printed but the program in not stopped (and err is negative).\n\
@item 'stop'\n\
An error message is printed and the program is stopped.\n\
@end table\n\
\n\
dreamrect_f_p is an oct-function that is a part of the DREAM Toolbox available at\n\
@url{http://www.signal.uu.se/Toolbox/dream/}.\n\
\n\
Copyright @copyright{} 2006-2008 Fredrik Lingvall.\n\
@seealso {dreamrect_f, dreamrect}\n\
@end deftypefn")
{
  double *ro,*geom_par,*s_par,*m_par;
  dream_idx_type nt, no;
  int    ifoc=0;
  char   foc_met[50];
  int    buflen;
  double a, b, dx, dy, dt;
  double *delay,v,cp,alfa,focal=0;     
  double *h, *err_p;
  int    err_level=STOP, is_set = FALSE, err;
  char   err_str[50];
  DATA   *D;
  dream_idx_type    start, stop;
  pthread_t *threads;
  int    thread_n, N;
  void   *retval;
  sighandler_t   old_handler, old_handler_abrt, old_handler_keyint;
  octave_value_list oct_retval; 

  int nrhs = args.length ();  

  // Check for proper number of arguments
  
  if (!((nrhs == 8) || (nrhs == 9))) {
    error("dreamrect_f_p requires 8 or 9 input arguments!");
    return oct_retval; 
  }
  else
    if (nlhs > 2) {
      error("Too many output arguments for dreamrect_f_p !");
      return oct_retval; 
    }
  
  //
  // Observation point.
  //
  
  // Check that arg (number of observation points) x 3 matrix
  if (!mxGetN(0)==3) {
    error("Argument 1 must be a (number of observation points) x 3 matrix!");
    return oct_retval; 
  }
  no = mxGetM(0); // Number of observation points.
  const Matrix tmp0 = args(0).matrix_value();
  ro = (double*) tmp0.fortran_vec();

  //if (no<2) {
  //  error("At least 2 observation points i needed for this function!\n Use the serial version for a single observation point.");
  //  return oct_retval; 
  //}

  //
  // Transducer geometry
  //

  // Check that arg 2 is a 2 element vector
  if (!((mxGetM(1)==2 && mxGetN(1)==1) || (mxGetM(1)==1 && mxGetN(1)==2))) {
    error("Argument 2 must be a vector of length 2!");
    return oct_retval;
  }
  const Matrix tmp1 = args(1).matrix_value();
  geom_par = (double*) tmp1.fortran_vec();
  a = geom_par[0];		// x-width.
  b = geom_par[1];		// y-width.
  
  //
  // Temporal and spatial sampling parameters.
  //
  
  // Check that arg 3 is a 4 element vector
  if (!((mxGetM(2)==4 && mxGetN(2)==1) || (mxGetM(2)==1 && mxGetN(2)==4))) {
    error("Argument 3 must be a vector of length 4!");
    return oct_retval; 
  }
  const Matrix tmp2 = args(2).matrix_value();
  s_par = (double*) tmp2.fortran_vec();
  dx    = s_par[0];		// Spatial x-direction discretization size.
  dy    = s_par[1];		// Spatial y-direction discretization size.
  dt    = s_par[2];		// Temporal discretization size (= 1/sampling freq).
  nt    = (dream_idx_type) s_par[3];	// Length of SIR.

  //
  // Start point of impulse response vector ([us]). 
  //
  
  // Check that arg 4 is a scalar or vector.
  if ( (mxGetM(3) * mxGetN(3) !=1) && ((mxGetM(3) * mxGetN(3)) != no)) {
    error("Argument 4 must be a scalar or a vector with a length equal to the number of observation points!");
    return oct_retval; 
  }
  const Matrix tmp3 = args(3).matrix_value();
  delay = (double*) tmp3.fortran_vec();
  
  //
  // Material parameters
  //
  
  // Check that arg 5 is a 3 element vectora
  if (!((mxGetM(4)==3 && mxGetN(4)==1) || (mxGetM(4)==1 && mxGetN(4)==3))) {
    error("Argument 5 must be a vector of length 3!");
    return oct_retval; 
  }
  const Matrix tmp4 = args(4).matrix_value();
  m_par = (double*) tmp4.fortran_vec();
  v     = m_par[0]; // Normal velocity of transducer surface.
  cp    = m_par[1]; // Sound speed.
  alfa  = m_par[2]; // Attenuation coefficient [dB/(cm MHz)].

  //  
  // Focusing parameters.
  //
  
  //  ifoc = 1 - no foc, 2 foc x ,3 foc y, 4 foc xy (del=fsqrt(x*x+y*y)), 5 focx+focy.
  
  if (nrhs >= 6) {

    if (!mxIsChar(5)) {
      error("Argument 6 must be a string");  
      return oct_retval; 
    }

    std::string strin = args(5).string_value(); 
    buflen = strin.length();
    for (int n=0; n<=buflen; n++ ) {
      foc_met[n] = strin[n];
    }
    foc_met[buflen] = '\0'; 

    is_set = FALSE;
    
    if (!strcmp(foc_met,"off")) {
      ifoc = 1; 
      is_set = TRUE;
    }
    
    if (!strcmp(foc_met,"x")) {
      ifoc = 2; 
      is_set = TRUE;
    }

    if (!strcmp(foc_met,"y")) {
      ifoc = 3; 
      is_set = TRUE;
    }
    
    if (!strcmp(foc_met,"xy")) {
      ifoc = 4; 
      is_set = TRUE;
    }
    
    if (!strcmp(foc_met,"x+y")) {
      ifoc = 5; 
      is_set = TRUE;
    }

    if (is_set == FALSE) {
      error("Unknown focusing method!"); 
      return oct_retval; 
    }
    
    // Check that arg 7 is a scalar.
    if (mxGetM(6) * mxGetN(6) !=1 ) {
      error("Argument 7 must be a scalar!");
      return oct_retval; 
    }
    
    // Focal point (in mm).
    const Matrix tmp4 = args(6).matrix_value();
    focal = (double) tmp4.fortran_vec()[0];

  } else
    ifoc = 1;
  
  //
  // Number of threads.
  //

  // Check that arg 8 is a scalar.
  if ( mxGetM(7) * mxGetN(7) !=1) {
    error("Argument 8 must be a scalar!");
    return oct_retval; 
  }
  const Matrix tmp5 = args(7).matrix_value();
  N = (int) tmp5.fortran_vec()[0];
  if (N < 1) {
    error("Argument 8 must be larger or equal to 1 (min number of CPUs)!");
    return oct_retval; 
  }

  // n_cpus can't be larger then the number of observation points.
  if (N > no) { 
    N = no;
  }

  //  
  // Error reporting.
  //

  if (nrhs == 9) { 

    if (!mxIsChar(8)) {
      error("Argument 9 must be a string");
    return oct_retval; 
  }

    std::string strin = args(8).string_value(); 
    buflen = strin.length();
    for (int n=0; n<=buflen; n++ ) {
      err_str[n] = strin[n];
    }
    err_str[buflen] = '\0';     
    

    if (!strcmp(err_str,"ignore")) {
      err_level = IGNORE; 
      is_set = TRUE;
    }
    
    if (!strcmp(err_str,"warn")) {
      err_level = WARN; 
      is_set = TRUE;
    }
    
    if (!strcmp(err_str,"stop")) {
      err_level = STOP; 
      is_set = TRUE;
    }

    if (is_set == FALSE) {
      error("Unknown error level!");
      return oct_retval; 
    }
    
  }
  else
    err_level = STOP; // Default.


  // Create an output matrix for the impulse response
  Matrix h_mat(nt, no);
  h = h_mat.fortran_vec();

  //
  // Register signal handlers.
  //

  if ((old_handler = signal(SIGTERM, &sighandler)) == SIG_ERR) {
    printf("Couldn't register SIGTERM signal handler.\n");
  }

  if (( old_handler_abrt=signal(SIGABRT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register SIGABRT signal handler.\n");
  }
  
  if (( old_handler_keyint=signal(SIGINT, &sighandler)) == SIG_ERR) {
    printf("Couldn't register SIGINT signal handler.\n");
  }
  
  //
  // Call the DREAM subroutine.
  //

  out_err = NONE;
  running = TRUE;

#ifdef USE_FFTW
  if (alfa != (double) 0.0)
    att_init(nt,N);
#endif
  
  // Initialize the lock on out_err.
  pthread_mutex_init(&err_lock, NULL);
  
  // Allocate local data.
  D = (DATA*) malloc(N*sizeof(DATA));
  
  // Allocate mem for the threads.
  threads = (pthread_t*) malloc(N*sizeof(pthread_t));
  
  for (thread_n = 0; thread_n < N; thread_n++) {

    start = thread_n * no/N;
    stop =  (thread_n+1) * no/N;

    // Init local data.
    D[thread_n].start = start; // Local start index;
    D[thread_n].stop = stop; // Local stop index;
    D[thread_n].no = no;
    D[thread_n].ro = ro;
    D[thread_n].a = a;
    D[thread_n].b = b;
    D[thread_n].dx = dx;
    D[thread_n].dy = dy;
    D[thread_n].dt = dt; 
    D[thread_n].nt = nt;

    if (mxGetM(3) * mxGetN(3) == 1)
      D[thread_n].delay_method = SINGLE; // delay is a scalar.
    else
      D[thread_n].delay_method = MULTIPLE; // delay is a vector.

    D[thread_n].delay = delay;
    D[thread_n].v = v;
    D[thread_n].cp = cp;
    D[thread_n].alfa = alfa;
    D[thread_n].ifoc = ifoc;
    D[thread_n].focal = focal;
    D[thread_n].h = h;
    D[thread_n].err_level = err_level;
    
    // Starts the threads.
    err = pthread_create(&threads[thread_n], NULL, smp_process, &D[thread_n]);
    if (err != 0) {
      error("Error when creating a new thread!\n");
      return oct_retval; 
    }
  }
  
  // Wait for all threads to finish.
  for (thread_n = 0; thread_n < N; thread_n++) {
    err = pthread_join(threads[thread_n], &retval);
    if (err != 0) {
      error("Error when joining a thread!\n");
      return oct_retval; 
    }
  }

  // Free memory.
  free((void*) D);
 
  //
  // Restore old signal handlers.
  //

  if (signal(SIGTERM, old_handler) == SIG_ERR) {
    printf("Couldn't register old SIGTERM signal handler.\n");
  }
   
  if (signal(SIGABRT,  old_handler_abrt) == SIG_ERR) {
    printf("Couldn't register old SIGABRT signal handler.\n");
  }
  
  if (signal(SIGINT, old_handler_keyint) == SIG_ERR) {
    printf("Couldn't register old SIGINT signal handler.\n");
  }

#ifdef USE_FFTW
  if (alfa != (double) 0.0)
    att_close();
#endif 

  if (!running) {
    error("CTRL-C pressed!\n"); // Bail out.
    return oct_retval; 
  }

  //
  // Check for Error.
  //

  if ( (err_level == STOP) && (out_err != NONE)) {
    error(""); // Bail out if error.
    return oct_retval; 
  }

  oct_retval.append(h_mat);

  //  
  // Return error.
  //
  if (nlhs == 2) {
    Matrix err_mat(nt, no);
    err_p = err_mat.fortran_vec();
    err_p[0] = (double) out_err;
    oct_retval.append(err_mat);
  }
    
  return oct_retval;
}
